"""This module contains the debug script that can be used to start
the dummy experiment without the CLI.

"""
import os

from palaestrai.cli.manager import cli


def main():
    exp_path = os.path.abspath(
        os.path.join(
            __file__,
            "..",
            "..",
            "fixtures",
            "midas_experiment_run.yml",
        )
    )
    cli(["start", exp_path])


# def main(filename="dummy_experiment", verbose=0, use_custom_logging=False):
#     # runtime_file = os.path.abspath(
#     #     os.path.join(
#     #         __file__, "..", "..", "fixtures", "arl-runtime-debug.conf.yaml"
#     #     )
#     # )
#     # with open(runtime_file) as f:
#     #     conf = f.read()

#     RuntimeConfig().load()

#     init_logger(verbose)
#     if use_custom_logging:
#         custom_logging()

#     # experiment_file = os.path.abspath(
#     #     os.path.join(__file__, "..", "..", "fixtures", f"{filename}.yml")
#     # )
#     try:
#         experiment = Experiment.load(filename)
#     except (NotADirectoryError, FileNotFoundError):
#         experiment_file = os.path.abspath(
#             os.path.join(__file__, "..", "..", "fixtures", f"{filename}.yml")
#         )
#         experiment = Experiment.load(experiment_file)
#         print(
#             "Ignore that stupid, annoying error message. "
#             "I already found the file!!1."
#         )

#     executor = Executor()
#     executor.schedule(experiment)
#     executor_final_state = asyncio.run(executor.execute())
#     if executor_final_state != ExecutorState.EXITED:
#         sys.exit(
#             {
#                 ExecutorState.SIGINT: -2,
#                 ExecutorState.SIGABRT: -6,
#                 ExecutorState.SIGTERM: -15,
#             }[executor_final_state]
#         )


# def custom_logging():
#     """Manually configure palaestrai loggers."""
#     logging.getLogger("palaestrai").setLevel("ERROR")
#     logging.getLogger("palaestrai.agent").setLevel("ERROR")
#     logging.getLogger("palaestrai.core").setLevel("ERROR")
#     logging.getLogger("palaestrai.store").setLevel("ERROR")
#     logging.getLogger("palaestrai.experiment").setLevel("ERROR")
#     logging.getLogger("palaestrai.environment").setLevel("ERROR")
#     logging.getLogger("palaestrai.util").setLevel("ERROR")
#     logging.getLogger("palaestrai_mosaik").setLevel("DEBUG")
#     logging.getLogger("palaestrai_mosaik.environment").setLevel("ERROR")
#     logging.getLogger("palaestrai_mosaik.mosaikpatch").setLevel("ERROR")
#     logging.getLogger("palaestrai_mosaik.simulator").setLevel("ERROR")

#     logging.getLogger("pandapower").setLevel("ERROR")
#     logging.getLogger("matplotlib").setLevel("ERROR")
#     logging.getLogger("h5py").setLevel("ERROR")
#     logging.getLogger("midas").setLevel("DEBUG")
#     logging.getLogger("pysimmods").setLevel("ERROR")
#     logging.getLogger("numexpr").setLevel("ERROR")


if __name__ == "__main__":
    # main("dummy_experiment", use_custom_logging=True)
    main()



"""This module contains the :class:`MosaikEnvironment`, which
allows to run mosaik co-simulations with palaestrAI.

"""
from __future__ import annotations

import threading
from typing import Any, Dict, Optional, Union
import warnings

from numpy.random import RandomState
from palaestrai.environment.environment import Environment
from palaestrai.environment.environment_baseline import EnvironmentBaseline
from palaestrai.environment.environment_state import EnvironmentState
from palaestrai.types import SimTime
from palaestrai.util import seeding
from palaestrai.util.dynaloader import load_with_params

from . import LOG
from .environment.mosaik_world import MosaikWorld
from .environment.reward import Reward
from .simulator import ShutdownError


class MosaikEnvironment(Environment):
    """The mosaik environment for palaestrAI.

    This class implements the abstract class
    :class:`palaestrai.environment.Environment` and allows the
    interaction between ARL agents and a mosaik co-simulation.

    Parameters
    ----------
    connection: str
        The connection to the broker, required by the *Environment*
        class.
    uid: str
        The uid of this environment, required by the *Environment*
        class.
    seed: int
        A random seed provided by the experiment.
    module: str
        Full import path to a module or class that provides the two
        functions *description_func* and *instance_func*.
    description_func: str
        Name of a function of *module* that returns information objects
        of sensors and actuators.
    instance_func: str
        Name of a function of *module* that returns a mosaik world
        object that has not yet started.
    step_size: int
        Step size in seconds for the ARL synchronization simulator.
    params: Dict[str, Any], optional
        Additional params that will be passed to *description_func* and

    Attributes
    ----------
    mworld: :class:`.MosaikWorld`
        Creates and holds the mosaik world object.

    """

    def __init__(
        self,
        uid: str,
        broker_uri: str,
        seed: int,
        module: str,
        description_func: str,
        instance_func: str,
        end: Union[int, str],
        reward: Optional[Dict[str, Any]] = None,
        step_size: Optional[int] = None,
        arl_sync_freq: Optional[int] = None,
        silence_missing_input_connections_warning: bool = False,
        params: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(uid, broker_uri, seed)

        self._module_name: str = module
        self._description_func: str = description_func
        self._instance_func: str = instance_func
        if arl_sync_freq is None:
            warnings.warn(
                "Using 'step_size' instead of 'arl_sync_freq' is "
                "deprecated. 'step_size' will be removed in the future"
                ". This is just a renaming, the function stays the "
                "same.",
                UserWarning,
            )
            self._arl_sync_freq: int = step_size
        else:
            self._arl_sync_freq: int = arl_sync_freq

        self._end: Union[str, int] = end
        self._mosaik_params: Dict[str, Any] = {} if params is None else params
        self.rng: RandomState = seeding.np_random(self.seed)[0]

        self.mworld: MosaikWorld
        self._env_task: threading.Thread
        self._event_communicated: threading.Event
        self._event_stepped: threading.Event
        self._event_shutdown: threading.Event

        self._ctr_com: int = -1
        self._ctr_mos: int = -1
        self._shutdown: bool = False
        self._mosaik_finished: bool = False
        self.missing_input_connection_warning: bool = (
            not silence_missing_input_connections_warning
        )

        if reward is not None:
            self.reward: Reward = load_with_params(
                reward["name"], reward["params"]
            )

    def start_environment(self):
        self._event_communicated = threading.Event()
        self._event_stepped = threading.Event()
        self._event_shutdown = threading.Event()

        self._ctr_com = -1
        self._ctr_mos = -1
        self._shutdown = False
        self._mosaik_finished = False

        LOG.debug(
            "EnvironmentCOM %s(id=0x%x, uid=%s) is creating world object "
            "...",
            self.__class__,
            id(self),
            self.uid,
        )
        self.mworld = MosaikWorld(self)
        LOG.debug(
            "EnvironmentCOM %s(id=0x%x, uid=%s) is setting up simulators, "
            "connecting entities ...",
            self.__class__,
            id(self),
            self.uid,
        )
        self.mworld.setup(
            self._module_name,
            self._description_func,
            self._instance_func,
            self._end,
            self._arl_sync_freq,
            self._mosaik_params,
        )
        self.mworld.mifs.missing_input_connections_warning = (
            self.missing_input_connection_warning
        )
        LOG.debug(
            "EnvironmentCOM %s(id=0x%x, uid=%s) is starting mosaik "
            "simulation process ... ",
            self.__class__,
            id(self),
            self.uid,
        )
        self._env_task = threading.Thread(target=self.start)
        self._env_task.start()
        LOG.info(
            "EnvironmentCOM %s(id=0x%x, uid=%s) finished setup. "
            "Co-simulation is now running.",
            self.__class__,
            id(self),
            self.uid,
        )
        return EnvironmentBaseline(
            sensors_available=self.mifs.sensors,
            actuators_available=self.mifs.actuators,
            simtime=SimTime(
                simtime_ticks=0, simtime_timestamp=self.mworld.now_dt
            ),
        )

    def update(self, actuators):
        """Update the environment with new actuator values.

        On each call,  *actuators* contains new values, which will be
        injected into the mosaik data graph so that simulators can
        fetch these values without even noticing.

        This method implements the communication part of the
        synchronization between worker process and the mosaik process.

        Parameters
        ----------
        actuators : list[ActuatorInformation]
            A *list* of actuators with new values provided by the
            agents.

        Returns
        -------
        tuple
            A *tuple* containing a *list*, a *float*, and a *bool*
            The list contains sensors with updated values, the float
            represents a reward defining the current state of the
            environment, and the bool is False if the environment is
            still running and True otherwise.

        """
        LOG.debug(
            "EnvironmentCOM %s(id=0x%x, uid=%s) New update request.",
            self.__class__,
            id(self),
            self.uid,
        )

        if len(actuators) > 0:
            LOG.debug(
                "EnvironmentCOM %s(id=0x%x, uid=%s) Now updating "
                "actuators ...",
                self.__class__,
                id(self),
                self.uid,
            )
            self.mifs.update_actuators(actuators)
        else:
            LOG.debug(
                "EnvironmentCOM %s(id=0x%x, uid=%s) actuator list "
                "is empty!",
                self.__class__,
                id(self),
                self.uid,
            )

        self._ctr_com += 1
        if self._mosaik_finished:
            LOG.debug(
                "EnvironmentCOM %s(id=0x%x, uid=%s) noticed that "
                "mosaik finished.",
                self.__class__,
                id(self),
                self.uid,
            )
            sensors = self.mifs.get_sensor_readings()
            reward = self.reward(sensors, actuators)
            return EnvironmentState(
                sensor_information=sensors,
                rewards=reward,
                done=True,
                simtime=SimTime(
                    simtime_ticks=self.mworld.simtime,
                    simtime_timestamp=self.mworld.now_dt,
                ),
            )

        while self._ctr_mos <= self._ctr_com:
            if self._mosaik_finished:
                LOG.debug(
                    "EnvironmentCOM %s(id=0x%x, uid=%s) noticed that "
                    "mosaik finished.",
                    self.__class__,
                    id(self),
                    self.uid,
                )
                sensors = self.mifs.get_sensor_readings()
                reward = self.reward(sensors, actuators)
                return EnvironmentState(
                    sensor_information=sensors,
                    rewards=reward,
                    done=True,
                    simtime=SimTime(
                        simtime_ticks=self.mworld.simtime,
                        simtime_timestamp=self.mworld.now_dt,
                    ),
                )
            self._event_communicated.set()
            self._event_stepped.wait()
            self._event_stepped.clear()

        LOG.debug(
            "EnvironmentCOM %s(id=0x%x, uid=%s): Now mosaik should be "
            "waiting. Starting communication ...",
            self.__class__,
            id(self),
            self.uid,
        )
        sensors = self.mifs.get_sensor_readings()
        LOG.debug(
            "EnvironmentCOM %s(id=0x%x, uid=%s) received sensor "
            "readings %s from mosaik at time %d (%s)",
            self.__class__,
            id(self),
            self.uid,
            sensors,
            self.mworld.simtime,
            self.mworld.now_dt,
        )
        if not sensors:
            LOG.warning(
                "EnvironmentCOM %s(id=0x%x, uid=%s) received empty "
                "sensor readings. Something might went wrong in the "
                "environment. Reward calculation might fail!",
                self.__class__,
                id(self),
                self.uid,
            )
        try:
            reward = self.reward(sensors, actuators)
        except Exception as err:
            LOG.error(
                "EnvironmentCOM %s(id=0x%x, uid=%s): Reward calculation"
                " failed with error: %s",
                self.__class__,
                id(self),
                self.uid,
                err,
            )
            reward = []

        return EnvironmentState(
            sensor_information=sensors,
            rewards=reward,
            done=False,
            simtime=SimTime(
                simtime_ticks=self.mworld.simtime,
                simtime_timestamp=self.mworld.now_dt,
            ),
        )

    def sync(self, time):
        """Allow synchronization with the message worker.

        This method is called each time the :class:`.ARLSyncSimulator`
        steps. Updates the values in the sensor objects directly from
        the mosaik data graph.

        The actuator values will not be set here, since each simulator
        fetches its input data directly and independently from mosaik.

        This method implements the mosaik part of the synchronization
        between worker and mosaik process.

        Parameters
        ----------
        time : int
            The current simulation time (current step) is required to
            retrieve the correct data from the mosaik world.

        """
        LOG.debug(
            "EnvironmentMOS %s(id=0x%x, uid=%s) updated.",
            self.__class__,
            id(self),
            self.uid,
        )
        self.mworld.update_sim_time(time)
        self.mifs.trigger_sensors(self.mworld.world, time)

        self._ctr_mos += 1

        if self._shutdown:
            raise ShutdownError
        LOG.debug(
            "EnvironmentMOS %s(id=0x%x, uid=%s) signalling COM ...",
            self.__class__,
            id(self),
            self.uid,
        )
        while self._ctr_mos > self._ctr_com:
            self._event_stepped.set()
            if self._shutdown:
                raise ShutdownError
            self._event_communicated.wait()
            self._event_communicated.clear()

        LOG.debug(
            "EnvironmentMOS %s(id=0x%x, uid=%s): now com should be "
            "waiting. Start next step ...",
            self.__class__,
            id(self),
            self.uid,
        )

    def shutdown(self, reset=False):
        """Try to shutdown the environment.

        Returns
        -------
        bool
            True if the environment was shut down, False otherwise.

        """
        LOG.info(
            "EnvironmentCOM %s(id=0x%x, uid=%s) attempting to stop "
            "the co-simulation ...",
            self.__class__,
            id(self),
            self.uid,
        )
        self._shutdown = True
        self._event_communicated.set()

        self._env_task.join()

        self._event_shutdown.wait()
        LOG.debug(
            "EnvironmentCOM %s(id=0x%x, uid=%s) The counter for mosaik "
            "should be -1 if the shutdown was "
            "sucessful. Counter value is : %d",
            self.__class__,
            id(self),
            self.uid,
            self._ctr_mos,
        )
        self.is_terminal = not reset
        return self._mosaik_finished

    def start(self):
        """Start the simulation process.

        As soon as the mosaik simulation is finished (or interrupted),
        events are set to notify the worker that the simulation is
        finished.

        """
        self.mworld.start()
        self._ctr_mos = -1
        self._mosaik_finished = True
        self._event_shutdown.set()
        self._event_stepped.set()

    @property
    def mifs(self):
        """Return the mosaik interface component.

        Holds the sensors and actuators and handles the exchange of
        data between the communication worker and the mosaik simpy
        process.

        Returns
        -------
        :class:`.MosaikInterfaceComponent`

        """
        return self.mworld.mifs



import unittest
from unittest.mock import MagicMock, PropertyMock

import numpy as np
from palaestrai.agent import ActuatorInformation, SensorInformation
from palaestrai.types import Box
from palaestrai_mosaik.config import AGENT_FULL_ID
from palaestrai_mosaik.environment import MosaikInterfaceComponent
from palaestrai_mosaik.environment.mosaik_ifs import check_value, restore_value


class TestMosaikInterfaceComponent(unittest.TestCase):
    def setUp(self):
        self.sen_infos = [
            {
                "uid": "MySim-0.MyMod-0.attr",
                "space": (
                    "Box(low=0, high=1, shape=(1,), dtype=np.float32)"
                ),
            },
            SensorInformation(
                value=0,
                space=Box(0, 1, (1,), np.float32),
                uid="MySim-0.MyMod-1.attr",
            ),
            {
                "uid": "MySim-1.MyMod-0.attr",
                "space": (
                    "Box(low=0, high=1, shape=(1,), dtype=np.float32)"
                ),
            },
            SensorInformation(
                value=0,
                space=Box(0, 1, (1,), np.float32),
                uid="MySim-1.MyMod-1.attr",
            ),
        ]
        self.act_infos = [
            ActuatorInformation(
                value=None,
                space=Box(0, 1, (1,), np.float32),
                uid="MySim-0.MyMod-0.attr2",
            ),
            {
                "uid": "MySim-0.MyMod-1.attr2",
                "space": (
                    "Box(low=0, high=1, shape=(1,), dtype=np.float32)"
                ),
            },
            {
                "uid": "MySim-1.MyMod-0.attr2",
                "space": (
                    "Box(low=0, high=1, shape=(1,), dtype=np.float32)"
                ),
            },
        ]

        self.new_acts = [
            ActuatorInformation(
                value=0.2,
                space=Box(0, 1, (1,), np.float32),
                uid="MySim-0.MyMod-0.attr2",
            ),
            ActuatorInformation(
                value=0.2,
                space=Box(0, 1, (1,), np.float32),
                uid="MySim-0.MyMod-1.attr2",
            ),
            ActuatorInformation(
                value=0.2,
                space=Box(0, 1, (1,), np.float32),
                uid="MySim-1.MyMod-0.attr2",
            ),
            ActuatorInformation(
                value=0.4,
                space=Box(0, 1, (1,), np.float32),
                uid="MySim-0.MyMod-0.attr2",
            ),
        ]

    def test_create_sensors(self):
        mifs = MosaikInterfaceComponent(self.sen_infos, self.act_infos, 0)

        sensors = mifs.create_sensors()

        self.assertIsInstance(sensors, list)
        self.assertEqual(4, len(sensors))
        for sen in sensors:
            self.assertIsInstance(sen, SensorInformation)
            self.assertIsInstance(sen.space, Box)
            self.assertTrue(sen.space.contains(sen.value))

    def test_trigger_sensors(self):
        mifs = MosaikInterfaceComponent(self.sen_infos, self.act_infos, 0)
        sensors = mifs.create_sensors()

        df_cache = {
            0: {
                "MySim-0": {
                    "MyMod-0": {"attr": 0.5},
                    "MyMod-1": {"attr": 0.5},
                },
                "MySim-1": {
                    "MyMod-0": {"attr": 0.5},
                    "MyMod-1": {"attr": 0.5},
                },
            }
        }
        world = MagicMock()
        type(world)._df_cache = PropertyMock(return_value=df_cache)

        mifs.trigger_sensors(world, 0)

        self.assertEqual(len(sensors), len(mifs.sensors))
        for new_sen in mifs.sensors:
            match = False
            for old_sen in sensors:
                if new_sen.uid == old_sen.uid:
                    match = True
                    self.assertNotEqual(
                        new_sen.value, old_sen.value
                    )
                    break
            self.assertTrue(match)

    def test_get_sensor_readings(self):
        mifs = MosaikInterfaceComponent(self.sen_infos, self.act_infos, 0)
        sensors = mifs.create_sensors()

        readings = mifs.get_sensor_readings()

        self.assertEqual(len(sensors), len(readings))
        for sen, reading in zip(sensors, readings):
            self.assertEqual(sen.uid, reading.uid)
            self.assertEqual(sen.value, reading.value)
            self.assertEqual(sen.space, reading.space)

    def test_create_actuators(self):
        mifs = MosaikInterfaceComponent(self.sen_infos, self.act_infos, 0)

        actuators = mifs.create_actuators()

        self.assertIsInstance(actuators, list)
        self.assertEqual(3, len(actuators))
        for act in actuators:
            self.assertIsInstance(act, ActuatorInformation)
            self.assertIsInstance(act.space, Box)
            self.assertIsNone(act.value)

    def test_update_actuators(self):
        mifs = MosaikInterfaceComponent(self.sen_infos, self.act_infos, 0)

        mifs.create_actuators()
        mifs.update_actuators(self.new_acts)

        expected = [0.2, 0.2, 0.2, 0.4]

        for idx, actuator in enumerate(mifs.actuators):
            self.assertEqual(expected[idx], actuator.value)

    def test_trigger_actuators(self):
        mifs = MosaikInterfaceComponent(self.sen_infos, self.act_infos, 0)

        mifs.create_actuators()
        mifs.update_actuators(self.new_acts)

        sim = MagicMock()
        type(sim).sid = PropertyMock(return_value="MySim-0")
        input_data = dict()
        with self.assertLogs("palaestrai_mosaik.environment", "DEBUG") as cm:
            mifs.trigger_actuators(sim, input_data)

        self.assertIn("more than one agent", cm.output[-1])

        for eid, attrs in input_data.items():
            self.assertIsInstance(attrs, dict)
            for attr, src_ids in attrs.items():
                self.assertIsInstance(src_ids, dict)
                for src_id, val in src_ids.items():
                    self.assertEqual(AGENT_FULL_ID, src_id)
                    self.assertEqual(0.2, val)  # Depends on the seed!

    def test_check_value(self):
        value = {
            "MyKey": "my_value",
            "My2ndKey": {
                "MySubKey": 0,
            },
        }

        checked = check_value(value)
        self.assertIsInstance(checked, str)

    def test_restore_value(self):
        value1 = '{"MyKey": "my_value", "My2ndKey": {"MySubKey": 0}}'
        value2 = "Just a string"
        restored1 = restore_value(value1)
        restored2 = restore_value(value2)
        self.assertIsInstance(restored1, dict)
        self.assertIsInstance(restored2, str)


if __name__ == "__main__":
    unittest.main()



"""This module contains the :class:`.MosaikInterfaceComponent`, which
cares for all the precious sensors and actuators.

"""
from __future__ import annotations

import json
import warnings
from copy import copy
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import numpy as np
from palaestrai.agent.actuator_information import ActuatorInformation
from palaestrai.agent.sensor_information import SensorInformation
from palaestrai.types import Space

from palaestrai_mosaik.config import AGENT_FULL_ID
from palaestrai_mosaik.environment import LOG

if TYPE_CHECKING:
    from mosaik import World


class MosaikInterfaceComponent:
    """The mosaik interface component stores all sensors and actuators.

    Parameters
    ----------

    sensors : list
        List of sensor descriptions. A sensor description should either
        be an instance of
        :class:`palaestrai.agent.sensor_information.SensorInformation`
        or a dict like::

            sensor = {
                # Unique identifier of the sensor containing the
                # Simulator ID, the Entity ID, and the attribute uid
                "uid": <sid.eid.attr>,
                # String description of an *palaestrai.types* object
                "space": "Box(low=0.0, high=1.2, "
                                     "shape=(1,), dtype=np.float32)",
            }

    actuators : list
        List of actuator descriptions. An actuator description should
        either be an instance of
        :class:`palaestrai.agent.actuator_information.ActuatorInformation`
        or a dict like::

            actuator = {
                # Unique identifier of the actuator containing the
                # Simulator ID, the Entity ID, and the attribute uid
                "uid": <sid.eid.attr>,
                # String description of an *palaestrai.types* object
                "space": "Box(low=0.0, high=1.2, "
                                "shape=(1,), dtype=np.float32)",
            }

    seed : int, optional
        A seed for the random number generator (rng). When more than
        one agent provides values for the same actuator, one of the
        values is randomly chosen.

    Attributes
    ----------
    sensors : list
        A *list* containing all :class:`~.SensorInformation` objects.
    actuators : list
        A *list* containing all :class:`~.ActuatorInformation` objects.
    sensor_map : dict
        A *dict* that stores the index of each sensor as value.
    actuator_map: dict
        A *dict* that stores the valid actuator ids as keys and the
        initial configuration of each actuator as value.

    """

    def __init__(
        self,
        sensors: List[Union[SensorInformation, Dict[str, Any]]],
        actuators: List[Union[ActuatorInformation, Dict[str, Any]]],
        seed: Optional[int] = None,
    ):
        self._sensor_defs: List[
            Union[SensorInformation, Dict[str, Any]]
        ] = sensors
        self._actuator_defs: List[
            Union[ActuatorInformation, Dict[str, Any]]
        ] = actuators

        self.sensors: List[SensorInformation] = []
        self.actuators: List[ActuatorInformation] = []
        self.sensor_map: dict = {}
        self.actuator_map: dict = {}

        self._rng: np.random.RandomState = np.random.RandomState(seed)

        self.missing_input_connections_warning: bool = True

    def create_sensors(self) -> List[SensorInformation]:
        """Create sensors from the sensor description.

        The description is provided during initialization.

        Returns
        -------
        list
            The *list* containing the created sensor objects.

        """
        sensors = []
        for sensor in self._sensor_defs:
            if isinstance(sensor, SensorInformation):
                sensors.append(sensor)
                uid = sensor.uid
            else:
                uid = str(sensor.get("uid", sensor.get("sensor_id", "Unnamed Sensor")))
                try:
                    sensors.append(
                        SensorInformation(
                            uid=uid,
                            space=Space.from_string(sensor.get("space", sensor.get("observation_space", None))),
                            value=sensor.get("value", 0),
                        )
                    )
                except RuntimeError:
                    LOG.exception(sensor)
                    raise
            self.sensor_map[uid] = copy(sensors[-1])
        self.sensors = sensors
        return self.sensors

    def create_actuators(self) -> List[ActuatorInformation]:
        """Create actuators from the actuator description.

        The description is provided during initialization.

        Returns
        -------
        list
            The *list* containing the created actuator objects.

        """
        actuators = []
        for actuator in self._actuator_defs:
            if isinstance(actuator, ActuatorInformation):
                actuators.append(actuator)
                uid = actuator.uid
            else:
                uid = str(actuator.get("uid", actuator.get("actuator_id", "Unnamed Actuator")))

                try:
                    actuators.append(
                        ActuatorInformation(
                            value=actuator.get(
                                "value", actuator.get("setpoint", None)
                            ),
                            uid=uid,
                            space=Space.from_string(
                                actuator.get("space", actuator.get("action_space", None))
                            ),
                        )
                    )
                except RuntimeError:
                    LOG.exception(actuator)
                    raise
            self.actuator_map[uid] = copy(actuators[-1])
        self.actuators = actuators
        return self.actuators

    def trigger_sensors(self, world: World, time: int):
        """Fill the sensors with fresh data from mosaik.

        This method is called from within the mosaik process.
        At each sync point (i.e., the :class:`.ARLSyncSimulator` is
        stepping), all current outputs of the simulators are written
        into the sensors.

        Parameters
        ----------
        world : :class:`mosaik.scenario.World`
            The current world object.
        time : int
            The current simulation time

        """
        self.sensors = list()
        for sid, entities in getattr(world, "_df_cache")[time].items():
            for eid, attrs in entities.items():
                for attr, val in attrs.items():
                    key = f"{sid}.{eid}.{attr}"

                    if key not in self.sensor_map:
                        continue

                    LOG.debug("Trying to access sensor %s.", key)
                    try:
                        sensor = copy(self.sensor_map[key])
                    except IndexError:
                        warnings.warn(
                            f"Sensor with id {key} not found in available "
                            "sensors. Skipping this one.",
                            UserWarning,
                        )
                        continue
                    # TODO: Check if sensor id is still correct
                    try:
                        val = check_value(val)
                    except TypeError:
                        LOG.warning(
                            "The type of value %s (%s) from sensor %s "
                            "could not be mapped to one of (int, float"
                            ", str, bool). Setting this value to None!",
                            val,
                            type(val),
                            key,
                        )
                        val = None
                    log_val = (
                        f"{val[:50]} ... string too long!"
                        if (
                            val is not None
                            and isinstance(val, str)
                            and len(val) > 100
                        )
                        else val
                    )

                    LOG.debug("Reading value %s from sensor %s.", log_val, key)
                    sensor.value = val
                    self.sensors.append(sensor)

    def get_sensor_readings(self):
        """Get all current sensor readings.

        This method is called from within the communication process.

        Returns
        -------
        list
            The *list* with all sensors and updated values.

        """
        return self.sensors

    def update_actuators(self, actuators: list):
        """Update the actuator values.

        This method is called from within the communication process.
        Removes all previous actuators from the list and adds all
        actuators from *actuators* to the list if they are present in
        the :attr:`actuator_map`.

        Parameters
        ----------
        actuators : list
            A *list* of :class:`.ActuatorInformation` objects.

        """
        self.actuators = []

        for actuator in actuators:
            if not isinstance(actuator, ActuatorInformation):
                warnings.warn(
                    f"Provided actuator ({type(actuator)}) is not an "
                    f"instance of {ActuatorInformation.__class__}."
                    "Skipping. ",
                    UserWarning,
                )
                continue

            if actuator.uid not in self.actuator_map:
                warnings.warn(
                    "Encountered unknown actuator "
                    f"{actuator.uid}. Skipping.",
                    UserWarning,
                )
                continue

            self.actuators.append(actuator)

    def trigger_actuators(self, sim, input_data):
        """Forward the actuator values to the specific simulators.

        This method is called from within the mosaik process.
        In contrast to the sensor calls, the actuators can only be
        triggered per simulator. Each simulator calls the mosaik
        scheduler to fetch its *input_data*. After the normal fetching
        process, the input_data dict is filled with inputs for the
        simulator *sim*.

        In this method, all actuators are checked wether they have
        "better" data or the specific input of the simulator. These
        values are updated and the old values will be overwritten. Then
        the actuators are resetted, i.e., setpoints (values) are set to None.

        The *input_data* is then returned to the scheduler, which
        passes them to the calling simulator.

        Notes
        -----
            It is possible that some simulators may rely on multiple
            inputs for a certain attribute and throw an error after the
            ARL injection, or filter for specific data provider and,
            therefore, ignore the ARL input. As soon as this case
            occurs, this method should be adapted accordingly.

        Parameters
        ----------
        sim : :class:`mosaik.scenario.ModelMock`
            The *ModelMock* of the simulator calling the
            :func:`.get_input_data` function of the mosaik scheduler.
        input_data : dict
            The currently gathered input data for this simulator.
            ARL tries to manipulate them.

        """

        for actuator in self.actuators:
            key = actuator.uid
            act_sid, act_eid, attr = key.split(".")
            if sim.sid != act_sid:
                continue
            LOG.debug("Trying to access actuator %s.", key)

            try:
                setpoint = restore_value(actuator.value)
            except TypeError:
                warnings.warn(
                    f"Could not restore value {actuator.value}"
                    f" ({type(actuator.value)}) of actuator {key}. "
                    "Skipping!",
                    UserWarning,
                )
                continue

            if setpoint is None:
                LOG.debug("The value of actuator %s is None. Skipping!", key)
                continue

            if attr not in input_data.setdefault(act_eid, dict()):
                input_data[act_eid][attr] = dict()
                if self.missing_input_connections_warning:
                    warnings.warn(
                        f"Attribute {attr} not found in the input_data"
                        " of this simulator. Most likely, there is no "
                        f"connection from any simulator to {act_sid}."
                        f"{act_eid}.{attr} Creating this entry may be "
                        "cause of some strange and unexpected behavior"
                        " ... You can silence this warning by setting "
                        "the key "
                        "'silence_missing_input_connections_warning' "
                        "to true in the environment params of your "
                        "experiment run file",
                        UserWarning,
                    )

            LOG.debug("Setting value %s to actuator %s.", setpoint, key)
            if AGENT_FULL_ID in input_data[act_eid][attr]:
                input_data[act_eid][attr][AGENT_FULL_ID].append(setpoint)
            else:
                input_data[act_eid][attr] = {AGENT_FULL_ID: [setpoint]}

        for model, attrs in input_data.items():
            for attr, src_ids in attrs.items():
                if AGENT_FULL_ID not in src_ids:
                    continue
                assert len(src_ids) == 1
                if len(src_ids[AGENT_FULL_ID]) > 1:
                    warnings.warn(
                        "It seems more than one agent already tried to"
                        f" manipulate {attr} of {model}: found values "
                        f"{src_ids[AGENT_FULL_ID]}. "
                        "May fate decide the outcome of this battle "
                        "...",
                        UserWarning,
                    )
                    setpoint = self._rng.choice(src_ids[AGENT_FULL_ID])
                else:
                    setpoint = src_ids[AGENT_FULL_ID][0]
                src_ids[AGENT_FULL_ID] = setpoint


def check_value(val):

    if isinstance(val, dict):
        # A dict! Let's dump it to json
        return str(json.dumps(val))
    if isinstance(val, np.ndarray):
        # We got a numpy array! This may be root of evil, unseen
        # errors. We convert it to a list if it has multiple values,
        # otherwise we unpack it.
        warnings.warn(
            f"Got numpy array {val} as value. It will be converted to a "
            "list or unpacked, depending on its length.",
            UserWarning,
        )
        if len(val) > 1:
            val = list(val)
        else:
            val = val[0]
    return val


def restore_value(val):
    if isinstance(val, str):
        try:
            return json.loads(val)
        except ValueError:
            # Just a normal string
            pass

    if isinstance(val, (list, np.ndarray)):
        if len(val) > 1:
            val = list(val)
        else:
            val = val[0]
    return val



"""This module contains a patch for mosaik's world object.

Needs to be updated when the mosaik API changes.

Tested with mosaik version 2.6.0.

"""
import types
from typing import Literal, Optional, Union

import networkx
import pkg_resources
from mosaik import util
from mosaik.scenario import logger, tqdm
from palaestrai_mosaik.mosaikpatch import LOG, scheduler


def modify_world(world):
    """Modify the world object for the use with palaestrAI.

    A new method :meth:`.trigger_actuators` is installed, which is
    called by the :func:`.get_input_data` function of the modified
    :mod:`~.scheduler`. Furthermore, the
    :func:`mosaik.scenario.World.run` method is overwritten by the
    modified :func:`~.run` version below.

    Parameters
    ----------
    world : :class:`mosaik.scenario.World`
        The *world* object to be modified.

    """
    LOG.debug("Installing 'trigger_actuators' method.")
    setattr(world, "trigger_actuators", None)
    LOG.debug("Modifying 'run' method.")
    setattr(world, "run", types.MethodType(run, world))


def run(
    self,
    until: int,
    rt_factor: Optional[float] = None,
    rt_strict: bool = False,
    print_progress: Union[bool, Literal["individual"]] = True,
    lazy_stepping: bool = True,
):
    """A modified version of the :func:`mosaik.scenario.World.run`
    method of mosaik's world object.

    This :func:`~.run` method works exactly like the original one
    except that a different scheduler is used. Indeed, it is nearly the
    same code but since a modified scheduler is imported, this specific
    scheduler is called instead.

    Furthermore, the prints are replaced by logging outputs.

    See the mosaik documentation
    https://mosaik.readthedocs.io/en/latest/api_reference/mosaik.scenario.html
    :func:`mosaik.scenario.World.run` for the regular features of this
    method.

    """
    if self.srv_sock is None:
        raise RuntimeError(
            "Simulation has already been run and can only "
            "be run once for a World instance."
        )

    # Check if a simulator is not connected to anything:
    for sid, deg in sorted(list(networkx.degree(self.df_graph))):
        if deg == 0:
            logger.warning("{sim_id} has no connections.", sim_id=sid)

    self.detect_unresolved_cycles()

    trigger_edges = [
        (u, v) for (u, v, w) in self.df_graph.edges.data(True) if w["trigger"]
    ]
    self.trigger_graph.add_edges_from(trigger_edges)

    self.cache_trigger_cycles()
    self.cache_dependencies()
    self.cache_related_sims()
    self.cache_triggering_ancestors()
    self.create_simulator_ranking()

    logger.info("Starting simulation.")
    # 11 is the length of "Total: 100%"
    max_sim_id_len = max(max(len(str(sid)) for sid in self.sims), 11)
    until_len = len(str(until))
    self.tqdm = tqdm(
        total=until,
        disable=not print_progress,
        colour="green",
        bar_format=(
            None
            if print_progress != "individual"
            else "Total:%s {percentage:3.0f}%% |{bar}| %s{elapsed}<{remaining}"
            % (" " * (max_sim_id_len - 11), "  " * until_len)
        ),
        unit="steps",
    )
    for sid, sim in self.sims.items():
        sim.tqdm = tqdm(
            total=until,
            desc=sid,
            bar_format="{desc:>%i} |{bar}| {n_fmt:>%i}/{total_fmt}{postfix:10}"
            % (max_sim_id_len, until_len),
            leave=False,
            disable=print_progress != "individual",
        )
    import mosaik._debug as dbg  # always import, enable when requested

    if self._debug:
        dbg.enable()
    success = False
    try:
        util.sync_process(
            scheduler.run(self, until, rt_factor, rt_strict, lazy_stepping),
            self,
        )
        success = True
    except KeyboardInterrupt:
        logger.info("Simulation canceled. Terminating ...")
    finally:
        for sid, sim in self.sims.items():
            sim.tqdm.close()
        self.tqdm.close()
        self.shutdown()
        if self._debug:
            dbg.disable()
        if success:
            logger.info("Simulation finished successfully.")


# def run(
#     self,
#     until,
#     rt_factor=None,
#     rt_strict=False,
#     print_progress=True,
#     lazy_stepping=True,
# ):
#     """A modified version of the :func:`mosaik.scenario.World.run`
#     method of mosaik's world object.

#     This :func:`~.run` method works exactly like the original one
#     except that a different scheduler is used. Indeed, it is nearly the
#     same code but since a modified scheduler is imported, this specific
#     scheduler is called instead.

#     Furthermore, the prints are replaced by logging outputs.

#     See the mosaik documentation
#     https://mosaik.readthedocs.io/en/latest/api_reference/mosaik.scenario.html
#     :func:`mosaik.scenario.World.run` for the regular features of this
#     method.

#     """
#     if self.srv_sock is None:
#         raise RuntimeError(
#             "Simulation has already been run and can only "
#             "be run once for a World instance."
#         )

#     # Check if a simulator is not connected to anything:
#     for sid, deg in sorted(list(networkx.degree(self.df_graph))):
#         if deg == 0:
#             LOG.warning("WARNING: %s has no connections.", sid)

#     # Get major mosaik version
#     mm_version = int(pkg_resources.require("mosaik")[0].version.split(".")[0])
#     if mm_version >= 3:
#         self.detect_unresolved_cycles()

#         trigger_edges = [
#             (u, v)
#             for (u, v, w) in self.df_graph.edges.data(True)
#             if w["trigger"]
#         ]
#         self.trigger_graph.add_edges_from(trigger_edges)

#         self.cache_trigger_cycles()
#         self.cache_dependencies()
#         self.cache_related_sims()
#         self.cache_triggering_ancestors()
#         self.create_simulator_ranking()

#     LOG.debug("Starting simulation.")
#     import mosaik._debug as dbg  # always import, enable when requested

#     if self._debug:
#         dbg.enable()
#     try:
#         if mm_version >= 3:
#             util.sync_process(
#                 scheduler.run(
#                     self,
#                     until,
#                     rt_factor,
#                     rt_strict,
#                     print_progress,
#                     lazy_stepping,
#                 ),
#                 self,
#             )
#         else:
#             util.sync_process(
#                 scheduler.run(
#                     self, until, rt_factor, rt_strict, print_progress
#                 ),
#                 self,
#             )
#         LOG.debug("Simulation finished successfully.")
#     except KeyboardInterrupt:
#         LOG.info("Simulation canceled. Terminating ...")
#     finally:
#         self.shutdown()
#         if self._debug:
#             dbg.disable()
